/**
*
* Anti collision system 1.0 main.
*
*
*/

// Library headers
#include <avr/io.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>
#include <avr/interrupt.h>
//#include <avr/wdt.h>
#include <inttypes.h>

// Firmware headers
#include "config.h"
#include "srf08/srf08.h"
#include "usart/usart0_driver.h"
#include "adc/adc.h"
#include "can/can_lib.h"
#include "sharp2d120x/sharp2d120x.h"


// General defines
#define ALL 0
#define TRUE 1
#define FALSE 0
#define IR 0
#define SONAR 1

// Sensor Board LEDs
#define LED_DDR DDRC
#define LED_PORT PORTC

// Power LED
#define PWR_LED_DDR DDRD
#define PWR_LED_PORT PORTD
#define PWR_LED PD7

// Board connected detection
#define BOARD_DETECT_DDR DDRA
#define BOARD_DETECT_PORT PORTA
#define BOARD_DETECT_PIN PINA
#define BOARD_1_DETECT PA5
#define BOARD_2_DETECT PA4
#define BOARD_3_DETECT PA2
#define BOARD_4_DETECT PA3


typedef struct led_state {
	uint8_t duty;		// start-up duty cycle (255 = 100%)
	uint8_t stop;		// stop with this duty-cycle value
	uint8_t up;			// up or down
	uint8_t stepsize;	// duty-cycle inc/dec each dt
	uint8_t dt;			// 1ms resoulutio = max 0,255s
	uint8_t cycles;		// How many times..
	//uint8_t ondelay;	// 10 ms resolution = max 2,55s
	//uint8_t offdelay;	// 10 ms resolution = max 2,55s
	uint8_t show_crash;	// Light up when collision
} led_state;


// Sensor board values for master board (1-4) and for slave board (4-8)
uint8_t sensor_board_rawvalue[MAX_SENSOR_BOARD_COUNT*2][2];
uint8_t sensor_board_average[MAX_SENSOR_BOARD_COUNT*2];
uint8_t sensor_board_threshold[MAX_SENSOR_BOARD_COUNT*2];
uint8_t sensor_board_light[MAX_SENSOR_BOARD_COUNT];	// Only needed for own board

uint8_t ledstates = 0xF;	// off
led_state board_led_state[MAX_SENSOR_BOARD_COUNT];



// i2c addresses
#ifdef MASTER
uint8_t sonar_i2c_addr[] = { BOARD1_SONAR_I2CADDR, BOARD2_SONAR_I2CADDR, BOARD3_SONAR_I2CADDR, BOARD4_SONAR_I2CADDR };
#else
uint8_t sendbuf[8];
uint8_t sonar_i2c_addr[] = { BOARD5_SONAR_I2CADDR, BOARD6_SONAR_I2CADDR, BOARD7_SONAR_I2CADDR, BOARD8_SONAR_I2CADDR };
#endif


// Indicates if a new slave board update is received.
volatile uint8_t slave_data_received;
uint8_t disable_slave = FALSE;

// Timer prototype
void init_timer1(void);

void led_show_collision(uint8_t i);
void flash_led_once(uint8_t i);
void flash_led(uint8_t i, uint8_t start, uint8_t stop, uint8_t up, uint8_t stepsize, uint8_t dt, uint8_t cycles);
void led_showoff();
void led_show_collision(uint8_t i);
void update_leds(int ticks);

/**
* Initialization.
*/
void init() {
	slave_data_received = 0;
	ledstates = 0x0F;	// off
	disable_slave = FALSE;

	// Board LEDs
	DDRC = 0x0F;	// Set LED port as output
	PORTC = ledstates; // Apply default value
	_delay_ms(250);_delay_ms(250);_delay_ms(250);_delay_ms(250);_delay_ms(250);_delay_ms(250);

	// Flash power on LED once if slave, twice if master.
	//LED is set again after init.
	DDRD |= (1 << PWR_LED);
	PORTD &= ~(1 << PWR_LED);
	_delay_ms(250);_delay_ms(250);_delay_ms(250);
	PORTD |= (1 << PWR_LED);
	#ifdef MASTER
	_delay_ms(250);_delay_ms(250);_delay_ms(250);_delay_ms(250);_delay_ms(250);_delay_ms(250);
	PORTD &= ~(1 << PWR_LED);
	_delay_ms(250);_delay_ms(250);_delay_ms(250);
	PORTD |= (1 << PWR_LED);
	#endif

	// Set up USART for debugging
	usart0_init(ATMEGA128_BAUD(57600));
	// Bind stdout to use uart0_putc on USART (printf is active after this point)
	fdevopen(usart0_putc, NULL);
	
	#ifdef MASTER
		printf("Master\n\r");
	#else
		printf("Slave\n\r");
	#endif
	
	printf("\n\n\n\n\n\n\n\n\n\rConnected..\n\r");

	// Set up CAN interface
	printf("Initializing CANbus.. ");
	init_can();
	printf("..OK\n\r");

	// Initialize PWM sonar/ultrasound
	//printf("Initializing PWM sonar.. ");
	//srf_init();
	//printf("..OK\n\r");

	// Initialize I2C-bus attached sonar (initializes i2c itself)
	printf("Initializing I2C sonar.. ");
	DDRD = 0x00;
	PORTD = 0xFF;
	//DDRD = 0xFF;
	srf08_init();
	/* set the I2C bit rate generator to 100 kb/s */
	TWSR = 0x00;
	TWBR  = 250;
	TWCR = _BV(TWEN);
	printf("..OK\n\r");

	// Initialize Sharp IR sensor (on ADC)
	printf("Initializing Sharp IR rangefinder..");
	sharp_ir_init();
	printf("..OK\n\r");

	// Set up DDR and pull up for board detection
	BOARD_DETECT_DDR &= ~((1 << BOARD_1_DETECT)|(1 << BOARD_2_DETECT)|(1 << BOARD_3_DETECT)|(1 << BOARD_4_DETECT)); // Make input
	BOARD_DETECT_PIN |= (1 << BOARD_1_DETECT)|(1 << BOARD_2_DETECT)|(1 << BOARD_3_DETECT)|(1 << BOARD_4_DETECT); // Pull ups on

	// Clear sensor value buffer
	for(int i=0; i<MAX_SENSOR_BOARD_COUNT*2; i++) {
		sensor_board_rawvalue[i][IR]=0;
		sensor_board_rawvalue[i][SONAR]=0;
		sensor_board_average[i] = 0;
		sensor_board_threshold[i] = DEFAULT_COLLISION_THRESHOLD_CM;
		if(i<4) {
			sensor_board_light[i] = TRUE;

			board_led_state[i].duty = 0;
			board_led_state[i].stop = 0;
			board_led_state[i].up = 0;
			board_led_state[i].stepsize = 0;
			board_led_state[i].dt = 0;
			board_led_state[i].cycles = 0;
			board_led_state[i].show_crash = TRUE;
		}
	}

	// LED-kontroll (soft-blink)
	init_timer1();

	// Enable power on LED
	DDRD |= (1 << PWR_LED);
	PORTD &= ~(1 << PWR_LED);

	// Flash all board leds:
	flash_led_once(ALL);

	// Enable interrupts
	sei();
}

/**
* Returns 1 if there is a sensor board on input i 0 if not.
*/
int detect_board(int i) {
	uint8_t status = ~BOARD_DETECT_PIN;

	switch(i) {
		case 0:
			return (status & (1 << BOARD_1_DETECT))>0;
		break;
		case 1:
			return (status & (1 << BOARD_2_DETECT))>0;
		break;
		case 2:
			return (status & (1 << BOARD_3_DETECT))>0;
		break;
		case 3:
			return (status & (1 << BOARD_4_DETECT))>0;
		break;
		default:
		return 0;
	}
}

/**
* Atmel appnote
*/
int init_can() {
	U8 txt_can_ana[]="======= UART in CAN ANALYZER =======";
	U8 txt_ech_mod[]="======== UART in ECHO MODE =========";

	U8 u8_temp;
	U8 buffer[8];
	st_cmd_t message;

	//- Pull-up on TxCAN & RxCAN one by one to use bit-addressing
    CAN_PORT_DIR &= ~(1<<CAN_INPUT_PIN );
    CAN_PORT_DIR &= ~(1<<CAN_OUTPUT_PIN);
    CAN_PORT_OUT |=  (1<<CAN_INPUT_PIN );
    CAN_PORT_OUT |=  (1<<CAN_OUTPUT_PIN);

	printf("Waiting for CAN msg (autobaud conf)..\n\r");
	// --- Init CAN (special AUTOBAUD)
        //- Wait until activity on RxCAN
    //while ((CAN_PORT_IN & (1<<CAN_INPUT_PIN)) != 0);
        //- Reset CAN peripheral
    Can_reset();
        //- Set CAN Bit-timming
    can_init((U16)CAN_BAUDRATE);        // c.f. macro in "can_drv.h"
        //- Set CAN Timer Prescaler
    //CANTCON = CANBT1;                   // Why not !
}


/**
* Sends a CAN message with given id and data from buffer pointer.
*/
int send_can_msg(int message_id, uint8_t *buffer, uint8_t buffer_length) {
	U8 u8_temp;
	st_cmd_t message;

	// Store interrupt status, disable interrupts.
	uint8_t sreg;
	sreg = SREG;
    cli();

	// --- Init Rx data
	message.pt_data = buffer;
	//for(int i=0; i<8; i++) buffer[i]=0;

	// --- Tx Command
	message.id.ext = message_id;   // Incrementation of ID to revome possible clashes
	message.dlc = buffer_length;
	message.cmd = CMD_TX_DATA;

	// Send the message:
	// --- Enable Tx
	while(can_cmd(&message) != CAN_CMD_ACCEPTED);
	// --- Wait for Tx completed

	while(1) {
		u8_temp = can_get_status(&message);
		if (u8_temp != CAN_STATUS_NOT_COMPLETED) break; // Out of while
	}

	// ---- Exit if CAN error(s)
	if (u8_temp == CAN_STATUS_ERROR) printf("CAN err %d\n\r", u8_temp); // Out of function

	// Set interrupt status back to original state.
	SREG = sreg;
	return 0;
}


/**
* Main method for anticollision node.
*/
int main(void) {
	uint8_t tmp = 0;

	// Initialization
	init();

	// Delay a bit..
/*
	for(int l=0;l<50;l++) {
		for(int k=0;k<10;k++){
			_delay_ms(100);
		}
	}
*/

	// Set address on i2c
	/*int8_t ret;
	uint8_t addr = 0;
	while((ret = srf08_setaddr(addr++, 0xE0)<0)) ;
		printf("Ret: %d addr %d\n\r", ret, addr);
	*/



	// Main loop
	printf("Starting up..\n\r");
	while(1) {

		// Read sensor values on own sensor boards
		for(int i=0;i<MAX_SENSOR_BOARD_COUNT;i++) {
			if(detect_board(i)==1) {
				tmp=0;
				// Ping with sonar i
				cli();
				srf08_ping(sonar_i2c_addr[i]>>1, RANGE_CM);

				// Read from IR i
				tmp = (uint8_t)read_sharp_ir(i);
				sensor_board_rawvalue[i][IR] = tmp;
				sei();

				// Fetch range from sonar i
				_delay_ms(2);                    // Recommended delay is 70 ms
				tmp=0;
				cli();
				srf08_range(sonar_i2c_addr[i]>>1, 0, &tmp); // Read first echo
				sensor_board_rawvalue[i][SONAR] = tmp;
				//IR sensor board 4 seems to be unstable
				sei();
				//printf("Value: %d, addr: %2X\n\r", tmp, sonar_i2c_addr[i]);
			}
			else {
				sensor_board_rawvalue[i][IR] = 0;
				sensor_board_rawvalue[i][SONAR] = 0;
				tmp=0;
			}
			sensor_board_rawvalue[i][IR] = sensor_board_rawvalue[i][SONAR];
		}


		// Collect sensor values from slave board (via CANbus) if master
		// or send sensor values to master board if slave
		#ifdef MASTER
			// Wait for fresh slave board data
			//sei();
			int8_t maxdelay = 200; // * 0,1ms
			while(disable_slave == FALSE && slave_data_received!=TRUE && maxdelay--) {_delay_ms(1); if(maxdelay==0) {disable_slave = TRUE; }}	// Put a timeout here...

		#else
			for(int i=0;i<(MAX_SENSOR_BOARD_COUNT);i++) {
				sendbuf[i*2] = sensor_board_rawvalue[i][IR];
				sendbuf[i*2+1] = sensor_board_rawvalue[i][SONAR];
			}
			send_can_msg(CANID_COLLISION_STATUS_UPDATE, sendbuf, 8);
		#endif
		_delay_ms(10);


		// Compute average values and check for collisions
		for(int i=0;i<(MAX_SENSOR_BOARD_COUNT*2);i++) {
			sensor_board_average[i] = (sensor_board_rawvalue[i][IR]+sensor_board_rawvalue[i][SONAR])/2;
			if(i==1 || i==2 || i==3 ){
				printf("Brd: %d: Average: %d \n\r", i+1, sensor_board_average[i]);
			}
			//_delay_ms(10); //Should not be here! Only for debugging...
		}

		// Check for collisions on all sensor boards
		uint8_t collision = 0;
		printf("*--------------------*\n\r");
		for(int i=0;i<(MAX_SENSOR_BOARD_COUNT*2);i++) {
			if(1){
				printf("Brd: %d: ir: %d, sonar: %d \n\r", i+1, sensor_board_rawvalue[i][IR],sensor_board_rawvalue[i][SONAR]);
			}
			//_delay_ms(10); //Should not be here! Only for debugging...

			if((sensor_board_rawvalue[i][IR]<=DEFAULT_COLLISION_THRESHOLD_CM) && (sensor_board_rawvalue[i][SONAR]<=DEFAULT_COLLISION_THRESHOLD_CM)) {
				// Collision detected
				if(sensor_board_average[i]>0) {
					collision = 1;
					if(i<4) led_show_collision(i+1);
				}
			}
		}

		#ifdef MASTER
		// Alert with CAN-message if any distance is below current threshold
		// on a sensorboards IR AND Sonar sensor.
		if(collision) send_can_msg(CANID_COLLISION_DETECTED, sensor_board_average, 8);
		#endif


		// Delay a few ms.
		//_delay_ms(100);_delay_ms(100);
	}

	return 0;
}


#define CPU_FREQ 16000000L  /* set to clock frequency in Hz */
#if CPU_FREQ == 16000000L
#define OCR_1MS 62*2
#endif
/*
 * Initialize timer1 to use the main crystal clock and the output
 * compare interrupt feature to generate an interrupt approximately
 * once per millisecond to use as a general purpose time base.
 */
void init_timer1(void)
{
  TCCR1A = 0;	// No external pin connection.
  TCCR1B  = (1 << WGM12)| (0 << CS12)| (1 << CS11) | (1 << CS10); /* CTC, prescale = 128 */
  //TIFR1  |= _BV(OCIE1A)|_BV(TOIE1);
  TIMSK1 |= _BV(TOIE1)|_BV(OCIE1A);         /* enable output compare interrupt */

  TCNT1  = 0;
  OCR1A   = 4;//*125;//62; //62,5 actually.//OCR_1MS;                     /* match in aprox 1 ms */
}




/*
 * millisecond counter interrupt vector
 */
ISR(TIMER1_COMPA_vect) {
	//uint8_t tmp;
	static uint16_t pwrstat = 255;
	static uint8_t up = 0;
	static uint16_t ticks = 0;
	if(ticks++>=255) ticks = 0;


	// Sensor board LEDs
	update_leds(ticks);

	// PWR led softblink
	if(ticks==0) {
		if(up) {
			pwrstat=pwrstat+2;
			if(pwrstat>=255) up=0;
		} else {
			pwrstat=pwrstat-2;
			if(pwrstat<=5) up=1;
		}
	}
	if(pwrstat > ticks) {
		PWR_LED_PORT |= (1 << PWR_LED);	// LED i off
	} else {
		PWR_LED_PORT &= ~(1 << PWR_LED);	// LED i on
	}


	// Do something fancy with the LEDs of a collision detecting board:
	/*
	if(send_warning_message > ls) {
		send_can_msg(CANID_COLLISION_DETECTED, &buffer, 4);
		LED_PORT &= 0xF0;
	} else if (send_warning_message < ls){
		LED_PORT |= 0xFF;
	}*/

	// Switch LEDs on if collision detected
/*
	for(int i=0;i<MAX_SENSOR_BOARD_COUNT;i++) {
		if(sensor_board_light[i] && (sensor_board_rawvalue[i][IR]<=sensor_board_threshold[i]) && (sensor_board_rawvalue[i][SONAR]<=sensor_board_threshold[i])) {
			LED_PORT &= ~(1 << i);	// LED i on
		} else {
			LED_PORT |= (1 << i);	// LED i off
		}
	}
*/

}


void led_showoff() {
	// Do some fancy stuff
}

void led_show_collision(uint8_t i) {
	//board_led_state[i].
	flash_led(i, 255, 0, FALSE, 1, 1, 1);
}

void update_leds(int ticks) {

	// Update status on update-eager leds
	for(int i=0;i<MAX_SENSOR_BOARD_COUNT;i++) {
		// Bare hvis tid for oppdatering, og vi har flere cycles igjen
		if(ticks == 0 && board_led_state[i].cycles) {
			uint8_t stepsize = board_led_state[i].stepsize;
			uint8_t duty = board_led_state[i].duty;
			if(board_led_state[i].up) {
				if((duty+stepsize)<255)
					board_led_state[i].duty = duty+stepsize;
				else
					board_led_state[i].up = FALSE;
			} else {
				if((duty - stepsize)>=0)
					board_led_state[i].duty = duty-stepsize;
				else { if(board_led_state[i].duty != 0)board_led_state[i].duty=0;
					board_led_state[i].up = TRUE;
				}
			}

			if(board_led_state[i].duty==board_led_state[i].stop) {
				board_led_state[i].cycles--;
				if(board_led_state[i].cycles==0)board_led_state[i].duty = board_led_state[i].stop;
			}
		}


		if(board_led_state[i].duty > ticks) {
		//if(sensor_board_light[i] && (sensor_board_average[i]*8 > ticks)) {
			LED_PORT &= ~(1 << i);	// LED i on
		} else {
			LED_PORT |= (1 << i);	// LED i off
		}
	}


}


void flash_led_once(uint8_t i) {
	flash_led(i, 0, 0, TRUE, 4, 1, 1);
}

void flash_led(uint8_t i, uint8_t start, uint8_t stop, uint8_t up, uint8_t stepsize, uint8_t dt, uint8_t cycles) {
	int lim;
	if(i==0) lim=MAX_SENSOR_BOARD_COUNT; else { lim=i; i--; }
	for(i;i<lim;i++) {
		board_led_state[i].duty = start;		// start-at duty-cycle
		board_led_state[i].stop = stop;			// Stop-at duty-cycle
		board_led_state[i].up = up;				// Up or down
		board_led_state[i].stepsize = stepsize;	// duty-cycle incrementation each dt
		//board_led_state[i].dec = dec;			// duty-cycle decrementation each dt
		board_led_state[i].dt = dt;				// 1ms resoulutio = max 0,255s
		board_led_state[i].cycles = cycles*2;	// 10 ms resolution = max 2,55s
		//board_led_state[i].offdelay = offdelay;	// 10 ms resolution = max 2,55s
	}
}


#define NUM_BYTES_DATA 8 // Number of data in the data field
#define CONMOB_DISABLE 0x00
#define CONMOB_ENABLE_TRANS (1 << CONMOB0)
#define CONMOB_ENABLE_RESEP (1 << CONMOB1)
#define CONMOB_ENABLE_FRAMEBUF_RESEP ((1 << CONMOB1)|(1 << CONMOB0))
uint8_t buf[8];
ISR(SIG_CAN_INTERRUPT1){
	unsigned int id;
	uint8_t tmp;

	// Receive data on CAN channel 0

	//PORTC = 0x55;
	//printf("ID: %d, Dat: %d\n\r", id, CANMSG);
	CANPAGE = (0 << 4);

	if(CANSTMOB << 6) {
		id = (((int)(CANIDT2))>>5) + (((int)(CANIDT1))<<3);
		//printf("CAN rcvd id: %2X\n\r", id);


		switch(id){
			case CANID_SET_COLLISION_THRESHOLDS:
				#ifdef MASTER
				// Using 4 least significant bits
				for(int i=0;i<4;i++) sensor_board_threshold[i] = CANMSG;
				#else
				// Using 4 most significant bits
				for(int i=0;i<4;i++) tmp = CANMSG;	// FFWD 4 bytes
				for(int i=0;i<4;i++) sensor_board_threshold[i] = CANMSG;
				#endif
			break;
			case CANID_ENABLE_SENSOR_LEDS:
				#ifdef MASTER
				// Using 4 least significant bits
				for(int i=0;i<4;i++) sensor_board_light[i] = CANMSG;
				#else
				// Using 4 most significant bits
				for(int i=0;i<4;i++) tmp = CANMSG;	// FFWD 4 bytes
				for(int i=0;i<4;i++) sensor_board_light[i] = CANMSG;
				#endif
			break;
			case CANID_FLASH_SENSOR_BOARD_LED:
				// First byte is 0-8 where 0 is "broadcast" to all leds
				for(int i = 0;i<8;i++) buf[i] = CANMSG;
				#ifdef MASTER
					if(buf[0]>=0 && buf[0]<=4) {
						flash_led(buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6]);
					}
				#else
					if(buf[0]==0 || (buf[0]>=5 && buf[0]<=8)) {
						flash_led(buf[0]==0?0:buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6]);
					}
				#endif
			break;
			case CANID_SET_INTERRUPT_REFRESH_INTERVAL:




			break;
			#ifdef MASTER
			case CANID_REQUEST_MEASUREMENTS:
				send_can_msg(CANID_RANGE_MEASUREMENTS , sensor_board_average, 8);
			break;
			case CANID_COLLISION_STATUS_UPDATE:
				for(int i=4;i<8;i++) {
					sensor_board_rawvalue[i][IR] = CANMSG;
					sensor_board_rawvalue[i][SONAR] = CANMSG;
					sensor_board_rawvalue[i][IR] = sensor_board_rawvalue[i][SONAR];
					/*if(i == 4){
						sensor_board_rawvalue[i][IR] = sensor_board_rawvalue[i][SONAR]; //IR tends to measure wrong in short distances
					}*/
				}
				slave_data_received = TRUE;
			break;
			#endif
			default:
				//printf("Unidentified CAN msg received.\n\r");
			break;
		}
	}




	CANPAGE = 0;//(0 << 4);
	CANSTMOB = 0x00;
	CANEN2 |= (1 << 0);
	CANCDMOB = NUM_BYTES_DATA;			// Reception 8 bytes
	CANCDMOB |= CONMOB_ENABLE_RESEP;	// Receive
	CANGIT = CANGIT;

}





