/*
 * $Id: srf08.c,v 1.1 2003/11/15 22:42:54 bsd Exp $
 *
 */

#include <compat/twi.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <avr/interrupt.h>

#include <util/delay.h>

#include "i2c.h"
#include "srf08/srf08.h"



#define CPU_FREQ 16000000L  /* set to clock frequency in Hz */

#if CPU_FREQ == 16000000L
#define OCR_1MS 62*2
#elif CPU_FREQ == 14745600L
#define OCR_1MS 115
#endif

volatile uint16_t ms_count;

/* 
 * millisecond counter interrupt vector 
 */
ISR(TIMER0_COMP_vect)
{
	//printf("Interrupt1\n\r");
	ms_count++;
}

/*
ISR(TIMER0_OVF_vect)
{
	printf("Interrupt2\n\r");
	ms_count++;
}*/


/*
 * ms_sleep() - delay for specified number of milliseconds
 */
void ms_sleep(uint16_t ms)
{
  TCNT0  = 0;
  ms_count = 0;
  while (ms_count != ms)
    ;
}


/*
 * Initialize timer0 to use the main crystal clock and the output
 * compare interrupt feature to generate an interrupt approximately
 * once per millisecond to use as a general purpose time base.
 */
void init_timer0(void)
{
  TCCR0A = 0;
  TIFR0  |= _BV(OCIE0A)|_BV(TOIE0);
  TIMSK0 |= _BV(TOIE0)|_BV(OCIE0A);         /* enable output compare interrupt */
  TCCR0A  = _BV(WGM01)|_BV(CS02)|_BV(CS00); /* CTC, prescale = 128 */
  TCNT0  = 0;
  OCR0A   = OCR_1MS;                     /* match in aprox 1 ms */
}


void srf08_init() {
	init_timer0();
}


/*
 * srf08_ping - initiate a ping to the SRF08 module.  Returns
 * immidiately, the caller must then wait the appropriate amount of
 * time before reading the echo results.  Use 'srf08_range()' to read
 * the echo results.  'device' is the 7 bit I2C address of the SRF08
 * module.  'cmd' is the command to execute.
 *
 * Returns 0 on success, or -1 if an I2C error occured.  */
int8_t srf08_ping(uint8_t device, uint8_t cmd)
{
  /* start condition */
  if (i2c_start(0x08, 1))
    return -1;

  /* address slave device, write */
  if (i2c_sla_rw(device, 0, TW_MT_SLA_ACK, 1))
    return -2;

  /* write address */
  if (i2c_data_tx(0x00, TW_MT_DATA_ACK, 1))
    return -3;

  /* write command */
  if (i2c_data_tx(cmd, TW_MT_DATA_ACK, 1))
    return -4;

  if (i2c_stop())
    return -5;

  return 0;
}


/*
 * srf08_range - Read echo results from the SRF08 module.
 *
 * 'device' should be the 7 bit I2C address of the unit.  'echo'
 * should contain which echo to read, 0=first echo, 1=second echo,
 * etc.  'range' should point to a 2 byte location to hold the 16 bit
 * echo result.
 *
 * Returns 0 on success, or -1 if an I2C error occured.
 */
int8_t srf08_range(uint8_t device, uint8_t echo, uint16_t * range)
{
  uint8_t vh, vl;
  uint8_t addr;

  /* start condition */
  if (i2c_start(0x08, 1))
    return -1;

  /* address slave device, write */
  if (i2c_sla_rw(device, 0, TW_MT_SLA_ACK, 1))
    return -2;

  addr = echo*2 + 2;

  /* write address */
  if (i2c_data_tx(addr, TW_MT_DATA_ACK, 1))
    return -3;

  /* repeated start condition */
  if (i2c_start(0x10, 1))
    return -4;

   /* address slave device, read */
  if (i2c_sla_rw(device, 1, TW_MR_SLA_ACK, 1))
    return -5;

  /* read high byte */
  if (i2c_data_rx(&vh, I2C_ACK, TW_MR_DATA_ACK, 1))
    return -6;

  /* read low byte */
  if (i2c_data_rx(&vl, I2C_NACK, TW_MR_DATA_NACK, 1))
    return -7;

  if (i2c_stop())
    return -8;

  *range = (vh << 8) | vl;

  return 0;
}





/**
* Set a new i2c-address on a SRF08 or SRF10 sonar module.
*
* If you have multiple SRF10 og SRF08 attached, make sure only one of them has
* the oldaddr I2C-address, so not multiple SRFs are given the same new address.
*
* New SRF10s are commonly shipped with 0xE0 as default address.
*
* Valid address range is 0xE0 - 0xFE
*
* If the command is successful, you'll see one long red flash and 0-15 short
* flashes from the SRF LED. (0 means address set to 0xE0).
*/
int8_t srf08_setaddr(uint8_t oldaddr, uint8_t newaddr) {
	// CHANGEADDR1

	/* start condition */
  if (i2c_start(0x08, 1))
    return -1;
  /* address slave device, write */
  if (i2c_sla_rw(oldaddr, 0, TW_MT_SLA_ACK, 1))
    return -2;
  /* write address */
  if (i2c_data_tx(SRF08_COMMAND, TW_MT_DATA_ACK, 1))
    return -3;
  /* write command */
  if (i2c_data_tx(CHANGEADDR1, TW_MT_DATA_ACK, 1))
    return -4;
  if (i2c_stop())
    return -5;


/* start condition */
  if (i2c_start(0x08, 1))
    return -1;
  /* address slave device, write */
  if (i2c_sla_rw(oldaddr, 0, TW_MT_SLA_ACK, 1))
    return -2;
  /* write address */
  if (i2c_data_tx(SRF08_COMMAND, TW_MT_DATA_ACK, 1))
    return -3;
  /* write command */
  if (i2c_data_tx(CHANGEADDR2, TW_MT_DATA_ACK, 1))
    return -4;
  if (i2c_stop())
    return -5;


/* start condition */
  if (i2c_start(0x08, 1))
    return -1;
  /* address slave device, write */
  if (i2c_sla_rw(oldaddr, 0, TW_MT_SLA_ACK, 1))
    return -2;
  /* write address */
  if (i2c_data_tx(SRF08_COMMAND, TW_MT_DATA_ACK, 1))
    return -3;
  /* write command */
  if (i2c_data_tx(CHANGEADDR3, TW_MT_DATA_ACK, 1))
    return -4;
  if (i2c_stop())
    return -5;


/* start condition */
  if (i2c_start(0x08, 1))
    return -1;
  /* address slave device, write */
  if (i2c_sla_rw(oldaddr, 0, TW_MT_SLA_ACK, 1))
    return -2;
  /* write address */
  if (i2c_data_tx(SRF08_COMMAND, TW_MT_DATA_ACK, 1))
    return -3;
  /* write command */
  if (i2c_data_tx(newaddr, TW_MT_DATA_ACK, 1))
    return -4;
  if (i2c_stop())
    return -5;

	printf("Address is now %2X \n\r", newaddr);

	return 0;
}

