/*
** ** ** ** ** **  Begin header  ** ** ** ** ** ** **
**
** Project : Eurobot 2007
** Purpose : Easy to use Posix Message Queue implementation.
** Filename : posix_message_queue.h
** Version : 1.0
** Date : 18.01.2007
**
** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **
**
** Copyright © 2007, Kristian M. Knausgård
** All rights reserved.
**
** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **

VERSION LOG:
-----------
Version: 1.0
Date: 18.01.2007
Author: Kristian M. Knausgård
Coauthors: <none>
Description : Original version.
Target platform: Linux / Posix

FILE FUNCTIONALITY DESCRIPTION:
-----------

Posix Message Queue (PMQ) functions for easy use
of Posix Message Queues. See example main at bottom
of posix_message_queue.c.


Compile with:
gcc posix_message_queue.c -o pmqtest -lrt -lstdc++

** ** ** ** ** ** ** End header ** ** ** ** ** ** ** **
*/
#ifndef POSIX_MESSAGE_QUEUE_H
#define POSIX_MESSAGE_QUEUE_H

#include <stdio.h>
#include <string.h>
#include <mqueue.h>
#include <sys/signal.h>


#define PMODE 0666
#define MAX_POSIX_MESSAGE_SIZE 40
#define MAX_POSIX_MESSAGE_QUEUE_LENGTH 20

/**
* Typedef the Posix Message Queue descriptor to be
* consistent with name convention..
*/
typedef mqd_t posix_message_queue;

/**
* Posix message type.
*/
typedef struct pmq_message {
	int to;
	int from;
	int major;							// Message type ID.
	int minor;							// Message subtype ID.
	int priority;						// Message priority.
	int length;							// Data length.
	char data[MAX_POSIX_MESSAGE_SIZE];	// Message data.
} pmq_message;


/**
* Open the Posix Message Queue with specified name,
* or create and open if it doesn't exist.
*/
posix_message_queue pmq_open(char *pmq_name);


/**
* Send the pmq_message to the PMQ with name pmq_name.
*
* Non-blocing.
*/
int pmq_send(pmq_message *pm, posix_message_queue pmq);


/**
* Receive a PMQ message from PMQ with name pmq_name, and place it in the
* given pmq_message struct.
*
* Blocking.
*/
int pmq_receive(pmq_message *pm, posix_message_queue pmq);


/**
* Add a function to be called when a new Posix Message
* arrives in an empty queue. The function needs to take
* int callback_id as parameter (to identify the callback).
*
* Only ONE function can be registered as callback function
* for a Posix Message Queue. And it needs to be re-set as
* callback function each time a notification is received.
*
* Ex: void my_callback_fun(int callback_id)
*/
int pmq_set_listener(void(*callbackfun)(int), int callback_id, posix_message_queue pmq);


/**
* Close the PMQ.
*/
int pmq_close(posix_message_queue pmq);


/**
* Destroy the PMQ with the specified name (USE WITH CAUTION).
*
* Unlinks the PMQ -- one call from one user is needed, and
* the whole queue is gone.
*/
int pmq_destroy(char *pmq_name);


#endif
