#ifndef ROBOT_MODEL_H
#define ROBOT_MODEL_H

#include <iostream>
#include <fstream>
#include <time.h>
#include <sys/time.h>
#include "mathutils.h"
#include "Sensor.h"
#include "observations.h"
#include "OdometryObservation.h"
#include "observation_structs.h"
#include "Controller.h"

#include "Coordinate.h"
#include "CoordinateTransform.h"
#include "posix_message_queue.h"
#include "MessageHandler.h"
#include "LOSGuidance.h"
//#include "LOSPathGuidance.h"


using namespace std;


#define TABLE_LENGTH_X 3.000	// m
#define TABLE_WIDTH_Y 2.100	// m

#define ROBOT_RADIUS 0.200	// m

#define DRIVE_WHEEL_DIAMETER 0.076// m
#define DRIVE_WHEEL_DISTANCE 0.350// m


#define ODOMETRY_GEAR_RATIO 1//23
#define ODOMETRY_PULSES_PER_REVOLUTION 2000


#define ODOMETRY_WHEEL_DIAMETER 0.060// m
#define ODOMETRY_WHEEL_DISTANCE 0.175// m

#define MAX_FORCE 0xFF

#define FILTER_LENGTH 4


struct robot_state_3d {
	// Position & heading (Local/TABLE)
	//CoordLocal position;
	//AngleEuler heading;
	double x_l;
	double y_l;
	double z_l;
	// Linear velocity (BODY)
	double u_b;
	double v_b;
	double w_b;
	// Force (BODY)
	double X_b;
	double Y_b;
	double Z_b;
	// Euler angles (TABLE)
	double phi_l;
	double theta_l;
	double psi_l;
	// Ang. velocity (BODY)
	double p_b;
	double q_b;
	double r_b;
	// Ang. moment (BODY)
	double K_b;
	double M_b;
	double N_b;
};


void handleCommands(pmq_message *pm);

void handleWPLocalCoordinates(pmq_message *pm);
void handleWPRobotCoordinates(pmq_message *pm);
void handleRobotToLocalRequest(pmq_message *pm);
void handleOdometryMessage(pmq_message *pm);
void handleSetSpeedLimit(pmq_message *pm);
void handleSetSpeed(pmq_message *pm);
void handleSetPosition(pmq_message *pm);
void handleBeaconMeasurement(pmq_message *pm);
void handleAdjustPosition(pmq_message *pm);
void handleBeaconMeasureRequest(pmq_message *pm);


static OdometryObservation *odometry;


class RobotModel// : Model
{
public:
	// Takes sample time and starting position as parameters
	RobotModel(double T, MessageHandler *messageHandler, MessageHandler *messageHandlerPlan, vector<Controller *> controllers);

	~RobotModel();


	// x-pos
	const double getX() { return x.x_l; }
	// y-pos
	const double getY() { return x.y_l; }
	// Yaw
	const double getPsi() { return x.psi_l; }

	// Stop model AND send stop command on posix message queue.
	void stop();

	// Update model
	void update();
	void addSensor(Sensor s);


	void setPosition(double x_l, double y_l, double psi_l);
	void setMaxForce(int left, int right);
	void setMaxSpeed(double u_b, double r_b);
	void setSpeed(double u_b, double r_b);

	int adjustLeft;
	int adjustRight;
	int adjustPosition;
	int measuring;


private:
	Observation *getNextMeasurement();
	void setMotorForce(int left, int right);

	MessageHandler *messageHandler;
	MessageHandler *messageHandlerPlan;
	double T;
	robot_state_3dof x;
	int maxForceLeft;
	int maxForceRight;

	double max_u_b;
	double max_r_b;
	double u_b;
	double r_b;



	double ringbuf[FILTER_LENGTH];

	timeval timeStamp;


	/**
	* State transition matrix:
	*/
	static double F[3][3];

	/**
	* Process noise coupling into system / disturbances:
	*
	* Remember to multiply with T in init/constructor...
	*/
	static double Gamma[3][3];

	Controller *surgeRateController;
	Controller *yawRateController;

};

#endif

