/*
** ** ** ** ** **  Begin header  ** ** ** ** ** ** **
**
** Project : Eurobot 2007
** Purpose : Easy to use Posix Message Queue implementation.
** Filename : posix_message_queue.c
** Version : 1.0
** Date : 18.01.2007
**
** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **
**
** Copyright © 2007, Kristian M. Knausgård
** All rights reserved.
**
** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **

VERSION LOG:
-----------
Version: 1.0
Date: 18.01.2007
Author: Kristian M. Knausgård
Coauthors: <none>
Description : Original version.
Target platform: Linux / Posix

FILE FUNCTIONALITY DESCRIPTION:
-----------

Posix Message Queue (PMQ) functions for easy use
of Posix Message Queues. See example main at bottom.


Compile with:
gcc posix_message_queue.c -o pmqtest -lrt -lstdc++

** ** ** ** ** ** ** End header ** ** ** ** ** ** ** **
*/


#include "posix_message_queue.h"


/**
* Open the Posix Message Queue with specified name,
* or create and open if it doesn't exist.
*/
posix_message_queue pmq_open(char *pmq_name) {
	posix_message_queue pmq;
	struct mq_attr attr;
	int open_flags;

	// Attributes
	attr.mq_maxmsg = MAX_POSIX_MESSAGE_QUEUE_LENGTH;
	attr.mq_msgsize = sizeof(pmq_message);
	attr.mq_flags   = 0;


	// IMPORTANT:
	//getrlimit() should be used here to check if the length/size is ok.

	// Set the flags for the open of the queue.
	// No O_NONBLOCK flag implies that the queue is blocking on send if queue is full.
	// O_CREAT means create queue if it doesn't exist.
	// Specify O_WRONLY if only writing is neccessary, O_RDWR if both read/write.
	open_flags = O_RDWR |O_CREAT|O_NONBLOCK;

	// Open the queue, and create it if it's not already created.
	pmq = (posix_message_queue)mq_open(pmq_name,open_flags,PMODE,&attr);
	if (pmq == -1) {
		printf("mq_open failed for %s. Are you root?", pmq_name);
		return -1;
	};
	
	// Success.
	return pmq;
}


/**
* Send the pmq_message to the PMQ with name pmq_name.
*
* Non-blocing.
*/
int pmq_send(pmq_message *pm, posix_message_queue pmq) {
	int status;
	status = mq_send((mqd_t)pmq, (char *)pm, sizeof(pmq_message), pm->priority);
    if (status == -1) {
//	printf("Failed to send posix message to %i, major: %i, minor: %i\n", pm->to, pm->major, pm->minor);        
//	perror("mq_send failed.");
		return -1;
	}

    //printf("Message sent.\n");

	// Success.
	return 0;
}


/**
* Receive a PMQ message from PMQ with name pmq_name, and place it in the
* given pmq_message struct.
*
* Blocking.
*/
int pmq_receive(pmq_message *pm, posix_message_queue pmq) {
	int num_bytes_received;
	num_bytes_received = mq_receive((mqd_t)pmq, (char *)pm, sizeof(pmq_message), 0);
    if (num_bytes_received == -1) {
        //perror("mq_receive failed (-1).");
		return -1;
    } else if(num_bytes_received != sizeof(pmq_message)) {
		perror("mq_receive failed (only parts received) -- this should not happen.");
		return -1;
	}
    
	// Success.
	return 0;
}


/**
* Add a function to be called when a new Posix Message
* arrives in an empty queue. The function needs to take
* int callback_id as parameter (to identify the callback).
*
* Only ONE function can be registered as callback function
* for a Posix Message Queue. And it needs to be re-set as
* callback function each time a notification is received.
*
* Ex: void my_callback_fun(int callback_id)
*/
int pmq_set_listener(void(*callbackfun)(int), int callback_id, posix_message_queue pmq) {
	struct sigevent se;

	// SIGEV_THREAD, SIGEV_NONE or SIGEV_SIGNAL
	se.sigev_notify = SIGEV_THREAD;			// Type of notification.
	//se.sigev_signo             			// Signal number. 
	se.sigev_value = (union sigval)callback_id;  			// Signal value ("union sigval" type). 
	se.sigev_notify_function = (void *)callbackfun;	// Notification function.

	// (sigev_notify_attributes = NULL implies PTHREAD_CREATE_DETACHED)
	se.sigev_notify_attributes = NULL;		// Notification attributes.

	return mq_notify((mqd_t)pmq, &se);
}


/**
* Close the PMQ.
*/
int pmq_close(posix_message_queue pmq) {
	return mq_close((mqd_t)pmq);
}


/**
* Destroy the PMQ with the specified name (USE WITH CAUTION).
*
* Unlinks the PMQ -- one call from one user is needed, and
* the whole queue is gone.
*/
int pmq_destroy(char *pmq_name) {
	return mq_unlink(pmq_name);
}





// Test functions:

/**
* Test listen function for async. notification.
*/
posix_message_queue pmq4callback;
void pmq_listen(int callback_id) {
	pmq_message pm;

	printf("Mottok callback (id = %d): ", callback_id);
	pmq_receive(&pm, pmq4callback);
	printf("Melding:\nMajor %d, minor %d, priority %d, datalength %d: %s\n", pm.major, pm.minor, pm.priority, pm.length, pm.data);

	// Re-enable listening:
	pmq_set_listener(pmq_listen, 4, pmq4callback);
}




/**
* Main for test purposes.
*
*
int main(int argc, char **argv) {
	posix_message_queue pmq;
	pmq_message pm;

	// Open queue.
	if(argc<=1) pmq_destroy("/testq");
	pmq = pmq_open("/testq");

	// Let pmq_listen use this shared variable for using the pmq.
	// Only used for async. notification.
	pmq4callback = pmq;

	// Sender or receiver:
	if(argc>1){
		pm.major = 1;
		pm.minor = 5;
		pm.priority = 0;
		pm.length = 5;
		strcpy(pm.data, "Test");
	
		pmq_send(&pm, pmq);
		printf("Sendte melding\n");
		pmq_close(pmq);		
	} else {
		
		// Sync. example (blocking).
		pm.major = 0;
		pm.minor = 0;
		pm.priority = 0;
		pm.length = 0;
		strcpy(pm.data, "Empty");

		printf("Venter på melding..\n");
		pmq_receive(&pm, pmq);
		printf("Mottok med major %d, minor %d, priority %d, datalength %d: %s\n", pm.major, pm.minor, pm.priority, pm.length, pm.data);
		pmq_close(pmq);
		pmq_destroy("/testq");

		

		
		// Async. example (callback):
		printf("Setter opp asynkron notifikasjon..\n");
		pmq_set_listener(pmq_listen, 4, pmq);
		printf("Venter på melding..\n");
		while(1) { sleep(2000); }
	}
}
*/



