#include "MessageHandler.h"


void (* routingTable[MAX_MAJOR_NUMBER][MAX_MINOR_NUMBER])(pmq_message *);

vector<MessageHandler *> messageHandlerFromId;
MessageHandler *log;

/**
* Creates a MessageHandler object that receives messages on receiveFromQueue and
* sends messages on sendToQueue. The debug ID is important, as it identifies the
* sender in the log system.
*
* Queue names (that is the files) must start with prefix / and be >=3 characters
* in length.
*
* Two modes of operation is available; interrupt driven and simply use of the
* handleAllMessagesNow() method.
*/
MessageHandler::MessageHandler(int debugId, string receiveQueueName, string sendQueueName, bool interruptDrivenOperation) throw(invalid_argument) {
	this->debugId = debugId;
	this->receiveQueueName = receiveQueueName;
	this->sendQueueName = sendQueueName;
	this->interruptDriven = interruptDrivenOperation;

	// The log variable is global, and is used for shifting out log info. A
	// reference to the last MessageHandler object that is instanciated is
	// actually stored, but the log queue is identical for all MessageHandlers.
	log = this;
	logQueueName = PMQ_LOGSYS;

	if(receiveQueueName[0] != '/' || sendQueueName[0] != '/') {
		throw invalid_argument("The supplied argument is not a valid file descriptor / channel. Must start with '/'");
	} else if(receiveQueueName.length()<3 || sendQueueName.length()<3) {
		throw  invalid_argument("File descriptor / channel name must have length>=3");
	}

	printf("Opening receive channel %s..\n", receiveQueueName.c_str());
	if(!(receiveQueueDesc = pmq_open((char *)receiveQueueName.c_str()))) {
		throw invalid_argument("Could not connect to Posix Message Queue (inbound).");
	}

	printf("Opening send channel %s..\n", sendQueueName.c_str());
	if(!(sendQueueDesc = pmq_open((char *)sendQueueName.c_str()))) {
		throw invalid_argument("Could not connect to Posix Message Queue (outbound).");
	}

	if(receiveQueueName == logQueueName) {
		logQueueDesc = receiveQueueDesc;
	} else {
		printf("Opening log channel %s..\n", logQueueName.c_str());
		if(!(logQueueDesc = pmq_open((char *)logQueueName.c_str()))) {
			throw invalid_argument("Could not connect to Posix Message Queue (inbound).");
		}
	}

	//connectionActive = true;
	pmq_message tmp;
	while(pmq_receive(&tmp, receiveQueueDesc)>=0);	// Empty queue on startup (should negotiate here instead..)

	// Add handler now when the queue is empty. The order of arguments to this
	// "if" is very important, check isInterruptDriven() first!
	if(isInterruptDriven() && !pmq_set_listener(handleMessages, handlerId, receiveQueueDesc)) {
		switch(errno) {
			case EBUSY:
				throw invalid_argument("Could not add MessageHandler.handleMessage(). Another function is already listening on the PMQ.");
			break;
		}
	}
	cout << "PMQ Receive Ready...\n";

	// Clear routing table
	for(int i=0;i<MAX_MAJOR_NUMBER;i++) {
		for(int j=0;j<MAX_MINOR_NUMBER;j++) {
			routingTable[i][j] = NULL;
		}
	}

	// Add to interrupt translation table	
	handlerId = messageHandlerFromId.size();
	messageHandlerFromId.push_back(this);

	//logSysOstream = new LogSysOstream(cout, sendDebugMessage); 	// Overriding coutrea
}

/**
* Returns true if interrupt driven operation is chosen.
*/
bool MessageHandler::isInterruptDriven() {
	return interruptDriven;
}

/**
* Receive and handle all messages sequentially NOW (in given priority).
*/
void MessageHandler::handleAllMessagesNow() {
	handleMessage();
}

/**
* Send a single string.
*/
void MessageHandler::say(string s) {
	debug_text_message dtm;

	s = s.substr(0,MAX_DEBUG_TEXT_LENGTH);
		// Build debug message.
	//debug_text_message dtm;
	dtm.systemId = log->getDebugId();
	dtm.length = s.length();
	strcpy(dtm.text, s.c_str());

	// Build posix message.	
	pmq_message pm;
	pm.major = MAJOR_SPEAK;
	pm.minor = MINOR_DEFAULT_VOICE;
	pm.priority = PRIORITY_LOW;
	pm.length = sizeof(debug_text_message);
	memcpy(pm.data, (char *)&dtm, sizeof(debug_text_message));
	pmq_send(&pm, log->logQueueDesc);
}

/**
* Send a single string.
*/
void MessageHandler::sendMessage(string s) {
	pmq_message pm;
	pm.major = MAJOR_TEXT_MESSAGE;
	//pm.minor = 5;
	pm.priority = PRIORITY_VERY_LOW;
	pm.length = s.length();
	strcpy(pm.data, s.c_str());
	pmq_send(&pm, sendQueueDesc);
}

/**
* Send a debug message. Can also be done using <<.
*/
void sendDebugMessage(const string s) {
	// Build debug message.
	debug_text_message dtm;
	dtm.systemId = log->getDebugId();
	dtm.length = s.length();
	strcpy(dtm.text, s.c_str());

	// Build posix message.	
	pmq_message pm;
	pm.major = MAJOR_STRUCT;
	pm.minor = MINOR_LOG_PLAIN_TEXT;
	pm.priority = PRIORITY_LOW;
	pm.length = sizeof(debug_text_message);
	memcpy(pm.data, (char *)&dtm, sizeof(debug_text_message));
	pmq_send(&pm, log->logQueueDesc);
}

/**
* Send a log message.
*/
void MessageHandler::sendLogMessage(char *s, int size, int minor, int priority) {
	pmq_message pm;
	pm.major = MAJOR_STRUCT;
	pm.minor = minor;
	pm.priority = priority;
	pm.length = size;
	memcpy(pm.data, s, size);
	pmq_send(&pm, logQueueDesc);
}

/**
* Send a general struct.
*/
void MessageHandler::sendMessage(char *s, int size, int minor, int priority) {
	pmq_message pm;
	pm.major = MAJOR_STRUCT;
	pm.minor = minor;
	pm.priority = priority;
	pm.length = size;
	memcpy(pm.data, s, size);
	pmq_send(&pm, sendQueueDesc);
}

/**
 * Send a command.
 */
void MessageHandler::sendCommand(int command, int priority) {
  pmq_message pm;
  pm.major = MAJOR_COMMAND;
  pm.minor = command;
  pm.priority = priority;
  pm.length = 0;
  //memcpy(pm.data, s, size);	// No data.
  pmq_send(&pm, sendQueueDesc);
}

/**
* Add a message handler function to handle a specific message id.
*
* Add major and minor 0 if all messages should be forwarded to the function.
*/
void MessageHandler::addSubMessageHandler(void(*handlingfun)(pmq_message *), int major, int minor) {
	if(routingTable[major][minor]) cout << "WARNING: Existing SubMessageHandler overwritten\n";
	if(major == 0 && minor == 0) {
		for(int i=0;i<MAX_MAJOR_NUMBER;i++) {
			for(int j=0;j<MAX_MINOR_NUMBER;j++) {
				routingTable[i][j] = handlingfun;
			}
		}		
	} else if(minor == 0) {
		for(int j=0;j<MAX_MINOR_NUMBER;j++) {
			routingTable[major][j] = handlingfun;
		}		
	} else {
		routingTable[major][minor] = handlingfun;
	}
}

/**
* Top level Posix Message receiver for this process. Delegates messages
* to correct handeling function.
*/
static void handleMessages(int i) {
	messageHandlerFromId.at(i)->handleMessage();	
}

/**
* Top level Posix Message receiver for this MessageQueue. Delegates sub messages
* to correct handeling function.
*/
void MessageHandler::handleMessage() {
	pmq_message pm;

	// Readd listener if interrupt driven operation.
	if(isInterruptDriven()) { pmq_set_listener(handleMessages, handlerId, receiveQueueDesc); }

	//cout << "PM received with " << i << ":\n";
	//printf("Melding:\nMajor %d, minor %d, priority %d, datalength %d: %s\n", pm.major, pm.minor, pm.priority, pm.length, pm.data);
	while(pmq_receive(&pm, receiveQueueDesc)>=0)	
	switch(pm.major) {
		case MAJOR_COMMAND:
			if(routingTable[MAJOR_COMMAND][pm.minor]) (routingTable[MAJOR_COMMAND][pm.minor])(&pm);	// Process through routing table.
			//cout << "Received a command: " << pm.minor << "\n";
		break;
		case MAJOR_TEXT_MESSAGE:
			cout << "Received text: " << pm.data << "\n";
		break;
		case MAJOR_STRUCT:
			//cout << "Received c++ object type: " << pm.minor << " -- routing:\n";
			if(routingTable[MAJOR_STRUCT][pm.minor]) (routingTable[MAJOR_STRUCT][pm.minor])(&pm);	// Process through routing table.
		break;
		case MAJOR_SPEAK:
			if(routingTable[MAJOR_SPEAK][pm.minor]) (routingTable[MAJOR_SPEAK][pm.minor])(&pm);	// Process through
		break;
		default:
			// NOTHING TO DO...
			cout << "Warning, unhandled MAJOR id" << endl;
		break;
	}
}

const MessageHandler& MessageHandler::operator << ( const std::string& s ) const {
	sendDebugMessage(s);
	return *this;
}

// Debug message shiftout int
const MessageHandler& MessageHandler::operator << ( const int i ) const {
	char buf[5];
	sprintf(buf, "%d", i);
	sendDebugMessage(buf);
	return *this;
}

// Debug message shiftout double
const MessageHandler& MessageHandler::operator << ( const double d ) const {
	char buf[30];
	sprintf(buf, "%f", d);
	sendDebugMessage(buf);
	return *this;
}

// Debug message shiftout char
const MessageHandler& MessageHandler::operator << ( const char c) const {
	char buf[1];
	buf[0] = c;
	sendDebugMessage(buf);
	return *this;
}




