#include "PinPointedRuntime.h"
#include "Controller.h"


void timerHandler(int x);

string configFileName;
string controllerFileName;



/**
* Initialize all subsystems.
*/
int initPossys(bool headless) {

	printf("Setting up message system...\n");

	// This node receives on /navsys and sends to /proxy
	messageHandler = new MessageHandler(NAVSYS_ID, PMQ_NAVSYS, PMQ_PROXY, true);
	messageHandlerPlan = new MessageHandler(NAVSYS_ID, PMQ_NAVSYS, PMQ_PLANSYS, true);
	printf("...OK\n");

	printf("Reading configuration file \"%s\"...", configFileName.c_str());
	NavsysXMLParser environmentXMLParser(configFileName);
	printf("...OK\n");

	printf("Parsing configuration file \"%s\"...", configFileName.c_str());
	if(environmentXMLParser.parse("feature")) {
		printf("...OK\n");
	}
	else {
		return -1;
	}

	printf("Extracting features from configuration file \"%s\"...", configFileName.c_str());
	vector<Feature> environmentFeatures = environmentXMLParser.extractFeatures();
	printf("...OK\n");

	printf("Building environment map (static entries + dynamic entries w/initial conditions)...");
	EnvironmentMap environmentMap(environmentFeatures);
	printf("...OK\n");

	printf("Reading controller configuration file \"%s\"...", controllerFileName.c_str());
	NavsysXMLParser controllerXMLParser(controllerFileName);
	printf("...OK\n");

	printf("Parsing configuration file \"%s\"...", controllerFileName.c_str());
	if(controllerXMLParser.parse("robot")) {
		printf("...OK\n");
	} else {
		return -1;
	}

	printf("Extracting controllers from configuration file \"%s\"...", controllerFileName.c_str());
	vector<Controller *> robotControllers = controllerXMLParser.extractControllers();
	printf("...OK\n");

	if(headless) {
		printf("NOTE: Running in headless mode!\n");
	}
	else {
		printf("!headless... Doing nothing!\n");
		printf("...OK\n");
	}

	printf("NOT Testing configured sensor systems...");
	//SelfTest startupTest = new SelfTest(sensors);
	//startupTest.run();	// Maybe show visual..
	printf("...OK\n");

	printf("NOT Testing configured sensor systems...");
	//SelfTest startupTest = new SelfTest(sensors);
	//startupTest.run();	// Maybe show visual..
	printf("...OK\n");

	printf("Setting up robot model (RobotModel)...");
	// Testing model compilation:
	robotModel = new RobotModel(0.010, messageHandler, messageHandlerPlan, robotControllers); // Using T=10ms sampling time.
	printf("...OK\n");

	printf("Adding message handlers...");
	// Command handlers
	messageHandler->addSubMessageHandler( handleCommands , MAJOR_COMMAND, MINOR_ALL );

	// Struct handlers
	messageHandler->addSubMessageHandler( handleOdometryMessage , MAJOR_STRUCT, MINOR_ODOMETRY_OBSERVATION );
	messageHandler->addSubMessageHandler( handleWPLocalCoordinates, MAJOR_STRUCT,MINOR_NAVSYS_SET_WAYPOINT);
	messageHandler->addSubMessageHandler( handleWPRobotCoordinates, MAJOR_STRUCT,MINOR_NAVSYS_SET_WAYPOINT_ROBOT_RELATIVE);
	messageHandler->addSubMessageHandler( handleRobotToLocalRequest, MAJOR_STRUCT,MINOR_ROBOT_TO_LOCAL);
	messageHandler->addSubMessageHandler( handleSetSpeedLimit, MAJOR_STRUCT,MINOR_NAVSYS_SET_SPEED_LIMIT);
	messageHandler->addSubMessageHandler( handleSetSpeed, MAJOR_STRUCT,MINOR_NAVSYS_SET_SPEED);
	messageHandler->addSubMessageHandler( handleSetPosition, MAJOR_STRUCT,MINOR_NAVSYS_SET_POSITION);
	messageHandler->addSubMessageHandler( handleAdjustPosition, MAJOR_STRUCT, MINOR_NAVSYS_ADJUST_POSITION);
	messageHandler->addSubMessageHandler( handleBeaconMeasurement, MAJOR_STRUCT,MINOR_NAVSYS_BEACON_MEASUREMENT);
	messageHandler->addSubMessageHandler( handleBeaconMeasureRequest, MAJOR_STRUCT,MINOR_BEACON_MEASUREMENT_REQUEST);
	printf("...OK\n");


	return true;
}


void catch_sigint(int sig_num){
	// Re-register for SIGINT
	signal(SIGINT, catch_sigint);
	printf("Ctrl + c \n");

	// Force model to stop:
	robotModel->stop();

	exit(1);
}


void catch_sighup(int sig_num){
	// Re-register for SIGINT
	signal(SIGHUP, catch_sighup);
	// And ignore the signal.
}


/**
* Start up test program for positioning system.
*/
int main(int argc, char **argv) {
	int runAsBgProcess;
	runAsBgProcess = 1; // Default is on and not off... last minute change..

	if(argc==2 && strlen(argv[1])>=3){
		if((argv[1][0]=='-') && (argv[1][1]=='b')) {
			runAsBgProcess = 1;
		}
	}
	else if(strlen(DEFAULT_CONFIG_FILE_NAME)<MAX_CONFIG_FILE_NAME_LENGTH) {
	}

	configFileName = DEFAULT_CONFIG_FILE_NAME;
	controllerFileName = ROBOT_PARAMETER_FILE_NAME;


	// Safe stop mode:
	signal(SIGINT, catch_sigint);

	// Background mode if parameter set
	if(runAsBgProcess) signal(SIGHUP, catch_sighup);	// catch hang up signal (= run as background)


    	printf("\n\n\
	**************************************************\n\
	* Starting up positioning system test program... *\n\
	**************************************************");
	printf("\n\n\n");






// Run initialization.
	if(!initPossys(true)) {
		printf("Initialization failed...\n");
		return -1;
	}



	// Start timer
	// Start main-loop and timer.
	// Set up timer for pretty deterministic interrupt
	// Remember SIGALRM might be used for sleep()
	signal(SIGALRM, timerHandler);		// Timer handler
	struct itimerval timer2;      		// Timer struct
	timer2.it_interval.tv_sec = 0;      // Reset value
	timer2.it_interval.tv_usec = (__suseconds_t)((1.0/(double)UPDATE_RATE)*1000000);	// Convert UPDATE_RATE from Hz to ms
	timer2.it_value.tv_sec = timer2.it_interval.tv_sec; 	// Initial value
	timer2.it_value.tv_usec = timer2.it_interval.tv_usec;   // Initial value
	setitimer(ITIMER_REAL, &timer2, NULL);

	while(1) {
		sleep(1000);
	}


    return 0;
}




/**
* Periodic tasks triggered with timer interrupt.
*/
void timerHandler(int t) {
	// Reset -- to get more than on interrupt
	robotModel->update();

	signal(SIGALRM, timerHandler);
}


